__kupfer_name__ = _("Documents")
__kupfer_sources__ = ("RecentsSource", "PlacesSource", "IgnoredApps", )
__kupfer_actions__ = ("Toggle", )
__kupfer_contents__ = ("ApplicationRecentsSource", )
__description__ = _("Recently used documents and bookmarked folders")
__version__ = "2017.3"
__author__ = ""

import operator
import functools
from os import path
import xdg.BaseDirectory as base

from gi.repository import Gio, Gtk

from kupfer.objects import Action, Source
from kupfer.objects import SourceLeaf, AppLeaf, FileLeaf, UrlLeaf
from kupfer import icons
from kupfer import launch
from kupfer import plugin_support
from kupfer.weaklib import gobject_connect_weakly

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key" : "max_days",
        "label": _("Max recent document days"),
        "type": int,
        "value": 28,
    },
    {
        "key" : "check_doc_exist",
        "label": _("Show only existing documents"),
        "type": bool,
        "value": True,

    }
)

ALIASES = {
    'libreoffice': 'soffice',
}

# Libreoffice doesn't separate them out, so we'll hack that in manually
SEPARATE_APPS = {
    'libreoffice': {
        '.doc': 'libreoffice-writer',
        '.docx': 'libreoffice-writer',
        '.odt': 'libreoffice-writer',
        '.ods': 'libreoffice-calc',
        '.xlsx': 'libreoffice-calc',
        '.csv': 'libreoffice-calc',
        '.odp': 'libreoffice-impress',
        '.ppt': 'libreoffice-impress',
        '.pptx': 'libreoffice-impress',
        '.odg': 'libreoffice-draw',
        '.odf': 'libreoffice-math',
        '.mml': 'libreoffice-math',
    }
}

def _file_path(uri):
    try:
        return Gio.File.new_for_uri(uri).get_path()
    except Exception:
        # FIXME: encoding errors in get_path raise generic Exception
        return None

class RecentsSource (Source):
    def __init__(self, name=None):
        if not name:
            name = _("Recent Items")
        super().__init__(name)

    def initialize(self):
        """Set up change callback"""
        manager = Gtk.RecentManager.get_default()
        gobject_connect_weakly(manager, "changed", self._recent_changed)

    def _recent_changed(self, *args):
        # FIXME: We don't get single item updates, might this be
        # too many updates?
        self._get.cache_clear()
        self.mark_for_update()

    def get_items(self):
        max_days = __kupfer_settings__["max_days"]
        return self._get_items(max_days)

    @functools.lru_cache(maxsize=1)
    def _get(max_days):
        manager = Gtk.RecentManager.get_default()
        def first_word(s):
            return s.split(None, 1)[0]

        items = manager.get_items()
        item_leaves = []
        check_doc_exist = __kupfer_settings__["check_doc_exist"]
        for item in items:
            day_age = item.get_age()
            if max_days >= 0 and day_age > max_days:
                continue
            if check_doc_exist and not item.exists():
                continue
            if item.get_private_hint():
                continue
            if not item.is_local():
                continue
            uri = item.get_uri()
            file_path = _file_path(uri)
            if not file_path:
                continue
            item_leaves.append((file_path, item.get_modified(), item))
        item_leaves.sort(key=operator.itemgetter(1), reverse=True)
        return item_leaves

    @staticmethod
    def _get_items(max_days, for_app_names=None):
        """
        for_app_names: set of candidate app names, or None.
        """
        def first_word(s):
            return s.split(None, 1)[0]

        for file_path, _modified, item in RecentsSource._get(max_days):
            if for_app_names:
                apps = item.get_applications()
                in_low_apps = any(A.lower() in for_app_names for A in apps)
                in_execs = any(first_word(item.get_application_info(A)[0])
                               in for_app_names for A in apps)
                if not in_low_apps and not in_execs:
                    continue

            if for_app_names:
                accept_item = True
                for app_id, sort_table in SEPARATE_APPS.items():
                    if app_id in for_app_names:
                        _, ext = path.splitext(file_path)
                        ext = ext.lower()
                        if ext in sort_table and sort_table[ext] not in for_app_names:
                            accept_item = False
                            break
                if not accept_item:
                    continue
            yield FileLeaf(file_path)

    def get_description(self):
        return _("Recently used documents")

    def get_icon_name(self):
        return "document-open-recent"
    def provides(self):
        yield FileLeaf
        yield UrlLeaf

class ApplicationRecentsSource (RecentsSource):
    def __init__(self, application):
        # TRANS: Recent Documents for application %s
        name = _("%s Documents") % str(application)
        super().__init__(name)
        self.application = application

    def repr_key(self):
        return self.application.repr_key()

    def get_items(self):
        app_names = self.app_names(self.application)
        max_days = __kupfer_settings__["max_days"]
        self.output_debug("Items for", app_names)
        items = self._get_items(max_days, app_names)
        return items

    # Cache doesn't need to be large to serve main purpose:
    # there will be many identical queries in a row
    @staticmethod
    @functools.lru_cache(maxsize=10)
    def has_items_for_application(names):
        max_days = __kupfer_settings__["max_days"]
        for item in RecentsSource._get_items(max_days, names):
            return True
        return False

    def get_gicon(self):
        return icons.ComposedIcon(self.get_icon_name(),
                self.application.get_icon())
    def get_description(self):
        return _("Recently used documents for %s") % str(self.application)

    @classmethod
    def decorates_type(cls):
        return AppLeaf

    @classmethod
    def decorate_item(cls, leaf):
        if IgnoredApps.contains(leaf):
            return None
        app_names = cls.app_names(leaf)
        if cls.has_items_for_application(app_names):
            return cls(leaf)
        return None

    @classmethod
    def app_names(cls, leaf):
        "Return a frozenset of names"
        svc = launch.GetApplicationsMatcherService()
        ids = [leaf.get_id()]
        app_name = svc.application_name(ids[0])
        if app_name is not None:
            ids.append(app_name.lower())
        ids.append(leaf.object.get_executable())
        for k, v in ALIASES.items():
            if k in ids:
                ids.append(v)
        return frozenset(ids)

class PlacesSource (Source):
    """
    Source for items from gtk bookmarks
    """
    def __init__(self):
        super(PlacesSource, self).__init__(_("Places"))
        self.places_file = None
        self._version = 2

    def initialize(self):
        self.places_file = path.join(base.xdg_config_home, "gtk-3.0", "bookmarks")

    def get_items(self):
        """
        gtk-bookmarks: each line has url and optional title
        file:///path/to/that.end [title]
        """
        if not path.exists(self.places_file):
            return ()
        return self._get_places(self.places_file)

    def _get_places(self, fileloc):
        for line in open(fileloc):
            if not line.strip():
                continue
            items = line.split(None, 1)
            uri = items[0]
            gfile = Gio.File.new_for_uri(uri)
            if len(items) > 1:
                title = items[1].strip()
            else:
                disp = gfile.get_parse_name()
                title = path.basename(disp)
            locpath = gfile.get_path()
            if locpath:
                yield FileLeaf(locpath, title)
            else:
                yield UrlLeaf(gfile.get_uri(), title)

    def get_description(self):
        return _("Bookmarked folders")
    def get_icon_name(self):
        return "system-file-manager"
    def provides(self):
        yield FileLeaf
        yield UrlLeaf

class IgnoredApps(Source):
    # This Source is invisibile and has no content
    # It exists just to store (through the config mechanism) the list of apps
    # we ignore for recent documents content decoration
    def __init__(self):
        super().__init__(_("Toggle Recent Documents"))
        # apps is a mapping: app id (str) -> empty dict
        self.apps = {}

    def config_save(self):
        return self.apps

    def config_save_name(self):
        return __name__

    def config_restore(self, state):
        self.apps = state

    def initialize(self):
        IgnoredApps.instance = self

    def finalize(self):
        del IgnoredApps.instance

    def get_items(self):
        return []

    def provides(self):
        return ()

    @classmethod
    def add(cls, app_leaf):
        cls.instance.apps[app_leaf.get_id()] = {}
        # FIXME: Semi-hack to refresh the content
        cls.instance.mark_for_update()

    @classmethod
    def remove(cls, app_leaf):
        cls.instance.apps.pop(app_leaf.get_id(), None)
        cls.instance.mark_for_update()

    @classmethod
    def contains(cls, app_leaf):
        return app_leaf.get_id() in cls.instance.apps

    def get_leaf_repr(self):
        return InvisibleSourceLeaf(self)

class Toggle(Action):
    rank_adjust = -5
    def __init__(self):
        super().__init__(_("Toggle Recent Documents"))

    def item_types(self):
        yield AppLeaf

    def valid_for_item(self, leaf):
        if IgnoredApps.contains(leaf):
            return True
        app_names = ApplicationRecentsSource.app_names(leaf)
        return ApplicationRecentsSource.has_items_for_application(app_names)

    def has_result(self):
        return True

    def activate(self, leaf):
        if IgnoredApps.contains(leaf):
            IgnoredApps.remove(leaf)
        else:
            IgnoredApps.add(leaf)
        # Neat trick: We return the leaf,
        # and that updates the decoration
        leaf._content_source = None
        return leaf

    def get_description(self):
        return _("Enable/disable listing recent documents in content for this application")

class InvisibleSourceLeaf (SourceLeaf):
    """Hack to hide this source"""
    def is_valid(self):
        return False



# -*- encoding: utf-8 -*-
"""
Kupfer's Hosts API

Main definition and *constructor* classes.
"""

from kupfer.obj.grouping import GroupingLeaf

__author__ = ("Ulrik Sverdrup <ulrik.sverdrup@gmail.com>, "
              "Karol Będkowski <karol.bedkowsk+gh@gmail.com>" )

HOST_NAME_KEY = "HOST_NAME"
HOST_ADDRESS_KEY = "HOST_ADDRESS"
HOST_SERVICE_NAME_KEY = "HOST_SERVICE_NAME"
HOST_SERVICE_PORT_KEY = "HOST_SERVICE_PORT"
HOST_SERVICE_USER_KEY = "HOST_SERVICE_USER"
HOST_SERVICE_PASS_KEY = "HOST_SERVICE_PASS"
HOST_SERVICE_REMOTE_PATH_KEY = "HOST_SERVICE_REMOTE_PATH"


class HostLeaf(GroupingLeaf):
    grouping_slots = (HOST_NAME_KEY, HOST_ADDRESS_KEY)

    def get_icon_name(self):
        return "computer"


class HostServiceLeaf(HostLeaf):
    ''' Leaf dedicated for well known services like ftp, ssh, vnc '''
    def __init__(self, name, address, service, description, port=None,
            user=None, password=None, slots=None):
        _slots = {
            HOST_NAME_KEY: name,
            HOST_ADDRESS_KEY: address,
            HOST_SERVICE_NAME_KEY: service,
            HOST_SERVICE_PORT_KEY: port,
            HOST_SERVICE_USER_KEY: user,
            HOST_SERVICE_PASS_KEY: password,
        }
        if slots:
            _slots.update(slots)
        HostLeaf.__init__(self, _slots, name or address)
        self._description = description

    def get_description(self):
        return self._description




# -*- coding: utf-8 -*-


__kupfer_name__ = _("Thunderbird")
__kupfer_sources__ = ("ContactsSource", )
__kupfer_actions__ = ("NewMailAction", "AttachToNewMail", "NewMailWithBody", )
__description__ = _("Thunderbird/Icedove Contacts and Actions")
__version__ = "2017.2"
__author__ = "Karol Będkowski <karol.bedkowski@gmail.com>, US"

from kupfer.objects import Action
from kupfer.objects import TextLeaf, UrlLeaf, RunnableLeaf, FileLeaf
from kupfer.obj.apps import AppLeafContentMixin
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer import utils, icons
from kupfer.obj.grouping import ToplevelGroupingSource
from kupfer.obj.contacts import ContactLeaf, EmailContact, email_from_leaf

from kupfer.plugin import thunderbird_support as support

"""
Changes:
    2012-03-15: Karol Będkowski
        + activate_multiple for new mail action
"""


class ComposeMail(RunnableLeaf):
    ''' Create new mail without recipient '''
    def __init__(self):
        RunnableLeaf.__init__(self, name=_("Compose New Email"))

    def run(self):
        if not utils.spawn_async(['thunderbird', '--compose']):
            utils.spawn_async(['icedove', '--compose'])

    def get_description(self):
        return _("Compose a new message in Thunderbird")

    def get_icon_name(self):
        return "mail-message-new"


class NewMailAction(Action):
    ''' Createn new mail to selected leaf (Contact or TextLeaf)'''
    def __init__(self):
        Action.__init__(self, _('Compose Email To'))

    def activate(self, leaf):
        self.activate_multiple((leaf, ))

    def activate_multiple(self, objects):
        recipients = ",".join(email_from_leaf(L) for L in objects)
        if not utils.spawn_async(['thunderbird', 'mailto:%s' % recipients]):
            utils.spawn_async(['icedove', 'mailto:%s' % recipients])

    def get_icon_name(self):
        return "mail-message-new"

    def item_types(self):
        yield ContactLeaf
        # we can enter email
        yield TextLeaf
        yield UrlLeaf

    def valid_for_item(self, item):
        return bool(email_from_leaf(item))


class AttachToNewMail(Action):
    def __init__(self):
        super().__init__(_("Attach in Email To..."))

    def activate(self, leaf, iobj,):
        self.activate_multiple((leaf, ), (iobj, ))

    def activate_multiple(self, objects, iobjects):
        attachments = ",".join(L.object for L in objects)
        recipients = ",".join(email_from_leaf(L) for L in iobjects)
        args = ["-compose", "to='%s',attachment='%s'" % (recipients, attachments)]
        if not utils.spawn_async(['thunderbird'] + args):
            utils.spawn_async(['icedove'] + args)

    def get_icon_name(self):
        return "mail-message-new"

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, leaf):
        return not leaf.is_dir()

    def requires_object(self):
        return True

    def object_types(self):
        yield ContactLeaf
        # we can enter email
        yield TextLeaf
        yield UrlLeaf

    def valid_object(self, iobj, for_item=None):
        return bool(email_from_leaf(iobj))

class NewMailWithBody(Action):
    def __init__(self):
        super().__init__(_("Compose Email With"))

    def activate(self, leaf):
        args = ["-compose", "body='%s'" % leaf.object]
        if not utils.spawn_async(['thunderbird'] + args):
            utils.spawn_async(['icedove'] + args)

    def get_icon_name(self):
        return "mail-message-new"

    def item_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Compose a new message using the text as body")

class ContactsSource(AppLeafContentMixin, ToplevelGroupingSource,
        FilesystemWatchMixin):
    appleaf_content_id = ('thunderbird', 'icedove')

    def __init__(self, name=_("Thunderbird Address Book")):
        ToplevelGroupingSource.__init__(self, name, "Contacts")
        self._version = 2

    def initialize(self):
        ToplevelGroupingSource.initialize(self)
        abook_dirs = list(support.get_addressbook_dirs())
        if abook_dirs:
            self.monitor_token = self.monitor_directories(*abook_dirs)

    def monitor_include_file(self, gfile):
        return gfile and (gfile.get_basename().endswith('.mab') \
                or gfile.get_basename() == 'localstore.rdf')

    def get_items(self):
        for name, email in support.get_contacts():
            yield EmailContact(email, name)

        yield ComposeMail()

    def should_sort_lexically(self):
        return True

    def get_description(self):
        return _("Contacts from Thunderbird Address Book")

    def get_gicon(self):
        return icons.get_gicon_with_fallbacks(None, ("thunderbird", "icedove"))

    def provides(self):
        yield ContactLeaf
        yield RunnableLeaf



__kupfer_name__ = _("APT")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = (
    "ShowPackageInfo",
    "SearchPackageName",
    "InstallPackage",
)
__description__ = _("Interface with the package manager APT")
__version__ = ""
__author__ = "Martin Koelewijn <martinkoelewijn@gmail.com>, Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import os
import subprocess

from kupfer.objects import Action, Source, Leaf
from kupfer.objects import TextLeaf
from kupfer import icons, kupferstring, task, uiutils, utils
from kupfer import plugin_support

__kupfer_settings__ = plugin_support.PluginSettings(
    {
        "key": "installation_method",
        "label": _("Installation method"),
        "type": str,
        "value": "gksu -- apt-get install --yes",
    },
)


class InfoTask(task.Task):
    def __init__(self, text):
        super(InfoTask, self).__init__()
        self.text = text
        self.aptitude = None
        self.apt_cache = None

    def start(self, finish_callback):
        self._finish_callback = finish_callback
        timeout = 60
        AC = utils.AsyncCommand
        AC(["apt", "show", self.text], self.aptitude_finished, timeout)
        AC(["apt-cache", "policy", self.text], self.aptcache_finished, timeout)

    def aptitude_finished(self, acommand, stdout, stderr):
        self.aptitude = stderr
        self.aptitude += stdout
        self._check_end()

    def aptcache_finished(self, acommand, stdout, stderr):
        self.apt_cache = stderr
        self.apt_cache += stdout
        self._check_end()

    def _check_end(self):
        if self.aptitude is not None and self.apt_cache is not None:
            self.finish("".join(kupferstring.fromlocale(s)
                                for s in (self.aptitude, self.apt_cache)))

    def finish(self, text):
        uiutils.show_text_result(text, title=_("Show Package Information"))
        self._finish_callback(self)


class ShowPackageInfo(Action):
    def __init__(self):
        Action.__init__(self, _("Show Package Information"))

    def is_async(self):
        return True

    def activate(self, leaf):
        return InfoTask(leaf.object.strip())

    def item_types(self):
        yield TextLeaf
        yield Package

    def valid_for_item(self, item):
        # check if it is a single word
        text = item.object
        return len(text.split(None, 1)) == 1

    def get_gicon(self):
        return icons.ComposedIcon("dialog-information", "package-x-generic")


class InstallPackage(Action):
    def __init__(self):
        Action.__init__(self, _("Install"))

    def activate(self, leaf):
        self.activate_multiple((leaf,))

    def activate_multiple(self, objs):
        program = (__kupfer_settings__["installation_method"])
        pkgs = [o.object.strip() for o in objs]
        prog_argv = utils.argv_for_commandline(program)
        utils.spawn_in_terminal(prog_argv + pkgs)

    def item_types(self):
        yield Package
        yield TextLeaf

    def get_description(self):
        return _("Install package using the configured method")

    def get_icon_name(self):
        return "document-save"


class Package(Leaf):
    def __init__(self, package, desc):
        Leaf.__init__(self, package, package)
        self.desc = desc

    def get_text_representation(self):
        return self.object

    def get_description(self):
        return self.desc

    def get_icon_name(self):
        return "package-x-generic"


class PackageSearchSource(Source):
    def __init__(self, query):
        self.query = query
        Source.__init__(self, _('Packages matching "%s"') % query)

    def repr_key(self):
        return self.query

    def get_items(self):
        package = kupferstring.tolocale(self.query)
        p = subprocess.run(['apt-cache', 'search', '--names-only', package],
                           capture_output=True)
        for line in kupferstring.fromlocale(p.stdout).splitlines():
            if not line.strip():
                continue
            if not " - " in line:
                self.output_error("apt-cache: ", line)
                continue
            package, desc = line.split(" - ", 1)
            yield Package(package, desc)

    def should_sort_lexically(self):
        return True

    def provides(self):
        yield TextLeaf

    def get_icon_name(self):
        return "system-software-install"


class SearchPackageName(Action):
    def __init__(self):
        Action.__init__(self, _("Search Package Name..."))

    def is_factory(self):
        return True

    def activate(self, leaf):
        package = leaf.object.strip()
        return PackageSearchSource(package)

    def item_types(self):
        yield TextLeaf

    def valid_for_item(self, item):
        # check if it is a single word
        text = item.object
        return len(text.split(None, 1)) == 1

    def get_icon_name(self):
        return "system-software-install"



# -*- coding: UTF-8 -*-
__kupfer_name__ = _("SSH Hosts")
__description__ = _("Adds the SSH hosts found in ~/.ssh/config.")
__version__ = "2010-04-12"
__author__ = "Fabian Carlström"

__kupfer_sources__ = ("SSHSource", )
__kupfer_actions__ = ("SSHConnect", )

import codecs
import os

from kupfer import icons, utils
from kupfer.objects import Action
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.obj.grouping import ToplevelGroupingSource
from kupfer.obj.hosts import HOST_NAME_KEY, HostLeaf, HOST_SERVICE_NAME_KEY, \
        HOST_ADDRESS_KEY



class SSHLeaf (HostLeaf):
    """The SSH host. It only stores the "Host" as it was
    specified in the ssh config.
    """
    def __init__(self, name):
        slots = {HOST_NAME_KEY: name, HOST_ADDRESS_KEY: name,
                HOST_SERVICE_NAME_KEY: "ssh"}
        HostLeaf.__init__(self, slots, name)

    def get_description(self):
        return _("SSH host")

    def get_gicon(self):
        return icons.ComposedIconSmall(self.get_icon_name(), "applications-internet")


class SSHConnect (Action):
    """Used to launch a terminal connecting to the specified
    SSH host.
    """
    def __init__(self):
        Action.__init__(self, name=_("Connect"))

    def activate(self, leaf):
        utils.spawn_in_terminal(["ssh", leaf[HOST_ADDRESS_KEY]])

    def get_description(self):
        return _("Connect to SSH host")

    def get_icon_name(self):
        return "network-server"

    def item_types(self):
        yield HostLeaf

    def valid_for_item(self, item):
        if item.check_key(HOST_SERVICE_NAME_KEY):
            return item[HOST_SERVICE_NAME_KEY] == 'ssh'
        return False


class SSHSource (ToplevelGroupingSource, FilesystemWatchMixin):
    """Reads ~/.ssh/config and creates leaves for the hosts found.
    """
    _ssh_home = os.path.expanduser("~/.ssh/")
    _ssh_config_file = "config"
    _config_path = os.path.join(_ssh_home, _ssh_config_file)

    def __init__(self, name=_("SSH Hosts")):
        ToplevelGroupingSource.__init__(self, name, "hosts")
        self._version = 2

    def initialize(self):
        ToplevelGroupingSource.initialize(self)
        self.monitor_token = self.monitor_directories(self._ssh_home)

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename() == self._ssh_config_file

    def get_items(self):
        try:
            content = codecs.open(self._config_path, "r", "UTF-8").readlines()
            for line in content:
                line = line.strip()
                words = line.split()
                # Take every word after "Host" as an individual host
                # we must skip entries with wildcards
                if words and words[0].lower() == "host":
                    for word in words[1:]:
                        if "*" in word:
                            continue
                        yield SSHLeaf(word)
        except EnvironmentError as exc:
            self.output_error(exc)
        except UnicodeError as exc:
            self.output_error("File %s not in expected encoding (UTF-8)" %
                    self._config_path)
            self.output_error(exc)

    def get_description(self):
        return _("SSH hosts as specified in ~/.ssh/config")

    def get_icon_name(self):
        return "applications-internet"

    def provides(self):
        yield SSHLeaf




import itertools
import os
from os import path

from gi.repository import GLib

from kupfer import datatools
from kupfer import icons
from kupfer import utils

from kupfer.obj.base import Source
from kupfer.obj.helplib import PicklingHelperMixin, FilesystemWatchMixin
from kupfer.obj.objects import FileLeaf, SourceLeaf
from kupfer.obj.objects import ConstructFileLeaf, ConstructFileLeafTypes

def _representable_fname(fname):
    "Return False if fname contains surrogate escapes"
    try:
        fname.encode("utf-8")
        return True
    except UnicodeEncodeError:
        return False

class FileSource (Source):
    def __init__(self, dirlist, depth=0):
        """
        @dirlist: Directories as byte strings
        """
        name = GLib.filename_display_basename(dirlist[0])
        if len(dirlist) > 1:
            name = _("%s et. al.") % name
        super(FileSource, self).__init__(name)
        self.dirlist = dirlist
        self.depth = depth

    def __repr__(self):
        return "%s.%s((%s, ), depth=%d)" % (self.__class__.__module__,
            self.__class__.__name__,
            ', '.join('"%s"' % d for d in sorted(self.dirlist)), self.depth)

    def get_items(self):
        iters = []
        
        def mkleaves(directory):
            files = utils.get_dirlist(directory, depth=self.depth,
                    exclude=self._exclude_file)
            return (ConstructFileLeaf(f) for f in files)

        for d in self.dirlist:
            iters.append(mkleaves(d))

        return itertools.chain(*iters)

    def should_sort_lexically(self):
        return True

    def _exclude_file(self, filename):
        return filename.startswith(".") 

    def get_description(self):
        return (_("Recursive source of %(dir)s, (%(levels)d levels)") %
                {"dir": self.name, "levels": self.depth})

    def get_icon_name(self):
        return "folder-saved-search"

    def provides(self):
        return ConstructFileLeafTypes()

class DirectorySource (Source, PicklingHelperMixin, FilesystemWatchMixin):
    def __init__(self, dir, show_hidden=False):
        # Use glib filename reading to make display name out of filenames
        # this function returns a `unicode` object
        name = GLib.filename_display_basename(dir)
        super(DirectorySource, self).__init__(name)
        self.directory = dir
        self.show_hidden = show_hidden

    def __repr__(self):
        return "%s.%s(\"%s\", show_hidden=%s)" % (self.__class__.__module__,
                self.__class__.__name__, str(self.directory), self.show_hidden)

    def initialize(self):
        self.monitor = self.monitor_directories(self.directory)

    def finalize(self):
        self.monitor = None

    def monitor_include_file(self, gfile):
        return self.show_hidden or not gfile.get_basename().startswith('.')

    def get_items(self):
        try:
            for fname in os.listdir(self.directory):
                if not _representable_fname(fname):
                    continue
                if self.show_hidden or not fname.startswith("."):
                    yield ConstructFileLeaf(path.join(self.directory, fname))
        except OSError as exc:
            self.output_error(exc)

    def should_sort_lexically(self):
        return True

    def _parent_path(self):
        return path.normpath(path.join(self.directory, path.pardir))

    def has_parent(self):
        return not path.samefile(self.directory , self._parent_path())

    def get_parent(self):
        if not self.has_parent():
            return super(DirectorySource, self).has_parent(self)
        return DirectorySource(self._parent_path())

    def get_description(self):
        return _("Directory source %s") % self.directory

    def get_gicon(self):
        return icons.get_gicon_for_file(self.directory)

    def get_icon_name(self):
        return "folder"

    def get_leaf_repr(self):
        if os.path.isdir(self.directory) and \
             os.path.samefile(self.directory, os.path.expanduser("~")):
            alias = _("Home Folder")
        else:
            alias = None
        return FileLeaf(self.directory, alias=alias)

    def provides(self):
        return ConstructFileLeafTypes()

class SourcesSource (Source):
    """ A source whose items are SourceLeaves for @source """
    def __init__(self, sources, name=None, use_reprs=True):
        if not name: name = _("Catalog Index")
        super(SourcesSource, self).__init__(name)
        self.sources = sources
        self.use_reprs = use_reprs

    def get_items(self):
        """Ask each Source for a Leaf substitute, else
        yield a SourceLeaf """
        for s in self.sources:
            yield (self.use_reprs and s.get_leaf_repr()) or SourceLeaf(s)

    def should_sort_lexically(self):
        return True

    def get_description(self):
        return _("An index of all available sources")

    def get_icon_name(self):
        return "kupfer-catalog"

class MultiSource (Source):
    """
    A source whose items are the combined items
    of all @sources
    """
    fallback_icon_name = "kupfer-catalog"
    def __init__(self, sources):
        super(MultiSource, self).__init__(_("Catalog"))
        self.sources = sources
    
    def is_dynamic(self):
        """
        MultiSource should be dynamic so some of its content
        also can be
        """
        return True

    def get_items(self):
        iterators = []
        ui = datatools.UniqueIterator(S.toplevel_source() for S in self.sources)
        for S in ui:
            it = S.get_leaves()
            iterators.append(it)

        return itertools.chain(*iterators)

    def get_description(self):
        return _("Root catalog")





__kupfer_name__ = _("Calculator")
__kupfer_actions__ = ("Calculate", )
__description__ = _("Calculate mathematical expressions")
__version__ = "2017.1"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

import locale
import cmath
import math

from kupfer.objects import Action, TextLeaf
from kupfer import pretty
import collections


class IgnoreResultException (Exception):
    pass


class KupferSurprise (float):
    """kupfer

    cleverness to the inf**inf
    """
    def __call__(self, *args):
        from kupfer import utils, version
        utils.show_url(version.WEBSITE)
        raise IgnoreResultException


class DummyResult (object):
    def __str__(self):
        return "<Result of last expression>"


class Help (object):
    """help()

    Show help about the calculator
    """
    def __call__(self):
        import textwrap
        import inspect

        from kupfer import uiutils

        environment = make_environment(last_result=DummyResult())
        docstrings = []
        for attr in sorted(environment):
            if attr != "_" and attr.startswith("_"):
                continue
            val = environment[attr]
            if not callable(val):
                docstrings.append("%s = %s" % (attr, val))
                continue
            try:
                try:
                    # use .replace() to remove unimportant '/' marker in signature
                    sig = str(inspect.signature(val)).replace(", /)", ")")
                    doc = "%s%s\n%s" % (attr, sig, val.__doc__)
                except ValueError:
                    doc = val.__doc__
                docstrings.append(doc)
            except AttributeError:
                pass
        formatted = []
        maxlen = 72
        left_margin = 4
        for docstr in docstrings:
            # Wrap the description and align continued lines
            docsplit = docstr.split("\n", 1)
            if len(docsplit) < 2:
                formatted.append(docstr)
                continue
            wrapped_lines = textwrap.wrap(docsplit[1].strip(),
                    maxlen - left_margin)
            wrapped = ("\n" + " " * left_margin).join(wrapped_lines)
            formatted.append("%s\n    %s" % (docsplit[0], wrapped))
        uiutils.show_text_result("\n\n".join(formatted), _("Calculator"))
        raise IgnoreResultException

    def __complex__(self):
        return self()


def make_environment(last_result=None):
    "Return a namespace for the calculator's expressions to be executed in."
    environment = dict(vars(math))
    environment.update(vars(cmath))
    # define some constants missing
    if last_result is not None:
        environment["_"] = last_result
    environment["help"] = Help()
    environment["kupfer"] = KupferSurprise("inf")
    # make the builtins inaccessible
    environment["__builtins__"] = {}
    return environment


def format_result(res):
    cres = complex(res)
    parts = []
    if cres.real:
        parts.append("%s" % cres.real)
    if cres.imag:
        parts.append("%s" % complex(0, cres.imag))
    return "+".join(parts) or "%s" % res


class Calculate (Action):
    # since it applies only to special queries, we can up the rank
    rank_adjust = 10
    # global last_result
    last_result = {'last': None}

    def __init__(self):
        Action.__init__(self, _("Calculate"))

    def has_result(self):
        return True

    def activate(self, leaf):
        expr = leaf.object.lstrip("= ")

        # try to add missing parantheses
        brackets_missing = expr.count("(") - expr.count(")")
        if brackets_missing > 0:
            expr += ")" * brackets_missing
        # hack: change all decimal points (according to current locale) to '.'
        #expr = expr.replace(locale.localeconv()['decimal_point'], '.')
        environment = make_environment(self.last_result['last'])
        pretty.print_debug(__name__, "Evaluating", repr(expr))
        try:
            result = eval(expr, environment)
            resultstr = format_result(result)
            self.last_result['last'] = result
        except IgnoreResultException:
            return
        except Exception as exc:
            pretty.print_error(__name__, type(exc).__name__, exc)
            resultstr = str(exc)
        return TextLeaf(resultstr)

    def item_types(self):
        yield TextLeaf

    def valid_for_item(self, leaf):
        text = leaf.object
        return text and text.startswith("=")

    def get_description(self):
        return None





"""
Tracker plugins are versioned by the D-Bus API version
This is version works with Tracker1.
"""
__kupfer_name__ = _("Tracker")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ("TrackerFulltext", )
__kupfer_contents__ = ("TrackerQuerySource", )
__kupfer_actions__ = (
        "TrackerSearch",
        "TrackerSearchHere",
    )
__description__ = _("Tracker desktop search integration")
__version__ = "2017.2"
__author__ = "US"

import os
from xml.etree.cElementTree import ElementTree

from gi.repository import Gio 
import dbus

from kupfer.objects import Action, Source
from kupfer.objects import TextLeaf, FileLeaf, TextSource, OperationError
from kupfer import utils, pretty
from kupfer import plugin_support


plugin_support.check_dbus_connection()

SERVICE_NAME = "org.freedesktop.Tracker"
SEARCH_OBJECT_PATH = "/org/freedesktop/Tracker/Search"
SEARCH_INTERFACE = "org.freedesktop.Tracker.Search"

SERVICE1_NAME = "org.freedesktop.Tracker1"
SEARCH_OBJECT1_PATH = "/org/freedesktop/Tracker1/Resources"
SEARCH1_INTERFACE = "org.freedesktop.Tracker1.Resources"

TRACKER_GUI_SEARCH = "tracker-needle"

class TrackerSearch (Action):
    def __init__(self):
        Action.__init__(self, _("Search in Tracker"))

    def activate(self, leaf):
        try:
            utils.spawn_async_raise([TRACKER_GUI_SEARCH, leaf.object])
        except utils.SpawnError as exc:
            raise OperationError(exc)
    def get_description(self):
        return _("Open Tracker Search Tool and search for this term")
    def get_icon_name(self):
        return "system-search"
    def item_types(self):
        yield TextLeaf

class TrackerSearchHere(Action):
    action_accelerator = "t"
    def __init__(self):
        super().__init__(_("Get Tracker Results..."))

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        query = leaf.object

        def error(exc):
            ctx.register_late_error(exc)

        def reply(results):
            ret = []
            new_file = Gio.File.new_for_uri
            for result in results:
                try:
                    ret.append(FileLeaf(new_file(result[0]).get_path()))
                except Exception: # This very vague exception is from getpath
                    continue
            ctx.register_late_result(TrackerQuerySource(query, search_results=ret))

        for _ignore in get_tracker_filequery(query, max_items=500,
                operation_err=True,
                reply_handler=reply, error_handler=error):
            pass

    def get_description(self):
        return _("Show Tracker results for query")
    def get_icon_name(self):
        return "tracker"
    def item_types(self):
        yield TextLeaf

def sparql_escape(ustr):
    """Escape unicode string @ustr for insertion into a SPARQL query

    Implemented to behave like tracker_sparql_escape in libtracker-client
    """
    sparql_escape_table = {
        ord('\t'): r'\t',
        ord('\n'): r'\n',
        ord('\r'): r'\r',
        ord('\b'): r'\b',
        ord('\f'): r'\f',
        ord('"') : r'\"',
        ord('\\'): '\\\\',
        # Extra rule: Can't have ?
        ord("?") : "",
    }
    return ustr.translate(sparql_escape_table)

ORDER_BY = {
    "rank": "ORDER BY DESC (fts:rank(?s))",
    "recent": "ORDER BY DESC (nfo:fileLastModified(?s))",
}
def get_file_results_sparql(searchobj, query, max_items=50, order_by="rank",
                            location=None, **kwargs):
    clean_query = sparql_escape(query)

    if location:
        location_filter = \
        'FILTER(tracker:uri-is-descendant ("%s", nie:url (?s)))' % sparql_escape(location)
    else:
        location_filter = ""

    sql = ("""SELECT tracker:coalesce (nie:url (?s), ?s)
              WHERE {
                ?s fts:match "%(query)s" .  ?s tracker:available true .
                %(location_filter)s
              }
              %(order_by)s
              OFFSET 0 LIMIT %(limit)d""" % dict(
              query=clean_query,
              location_filter=location_filter,
              order_by=ORDER_BY[order_by],
              limit=int(max_items))
              )

    pretty.print_debug(__name__, sql)
    results = searchobj.SparqlQuery(sql, **kwargs)
    if results is None:
        return

    new_file = Gio.File.new_for_uri
    for result in results:
        try:
            yield FileLeaf(new_file(result[0]).get_path())
        except Exception: # This very vague exception is from getpath
            continue

use_version = "0.8"
versions = {
    "0.8": (SERVICE1_NAME, SEARCH_OBJECT1_PATH, SEARCH1_INTERFACE),
}

version_query = {
    "0.8": get_file_results_sparql,
}


def get_searchobject(sname, opath, sinface, operation_err=False):
    bus = dbus.SessionBus()
    searchobj = None
    try:
        tobj = bus.get_object(sname, opath)
        searchobj = dbus.Interface(tobj, sinface)
    except dbus.DBusException as exc:
        if operation_err:
            raise OperationError(exc)
        pretty.print_debug(__name__, exc)
    return searchobj

def get_tracker_filequery(query, operation_err=False, **kwargs):
    searchobj = get_searchobject(*versions[use_version],
                                 operation_err=operation_err)
    if searchobj is None:
        pretty.print_error(__name__, "Could not connect to Tracker")
        return ()

    queryfunc = version_query[use_version]
    return queryfunc(searchobj, query, **kwargs)

class TrackerQuerySource (Source):
    def __init__(self, query, search_results=None, **search_args):
        Source.__init__(self, name=_('Tracker Search for "%s"') % query)
        self.query = query
        self.search_args = search_args
        self.search_results = None

    def repr_key(self):
        return self.query

    def get_items(self):
        if self.search_results:
            return self.search_results
        else:
            return get_tracker_filequery(self.query, **self.search_args)

    def provides(self):
        yield FileLeaf

    def get_description(self):
        return _('Results for "%s"') % self.query
    def get_icon_name(self):
        return "tracker"

    @classmethod
    def decorates_type(cls):
        return FileLeaf

    @classmethod
    def decorate_item(cls, leaf):
        # FIXME: Very simplified .savedSearch parsing, so far we only support
        # the query, without additional filtering. The simplest form of
        # .savedSearch file is saved by nautilus as following:
        # <query version="1.0">
        #   <text>QUERY GOES HERE</text>
        # </query>

        if not leaf.object.endswith(".savedSearch"):
            return None
        try:
            et = ElementTree(file=leaf.object)
            query = et.getroot().find("text").text
            if not query:
                return None
            location_tag = et.getroot().find("location")
            location = location_tag.text if location_tag is not None else None
            return cls(query, location=location_uri(location))
        except Exception:
            pretty.print_exc(__name__)
            return None

def location_uri(location):
    if location is None:
        return None
    if not os.path.isabs(location):
        location = os.path.expanduser("~/" + location)
    return Gio.File.new_for_path(location).get_uri()

# FIXME: Port tracker tag sources and actions
# to the new, much more powerful sparql + dbus API
# (using tracker-tag as in 0.6 is a plain hack and a dead end)

class TrackerFulltext (TextSource):
    def __init__(self):
        TextSource.__init__(self, name=_('Tracker Full Text Search'))

    def get_description(self):
        return _("Use '?' prefix to get full text results")

    def get_text_items(self, text):
        if text.startswith("?"):
            rank = "rank"
            if text.startswith("?~"):
                rank = "recent"
            query = text.lstrip("? ~")
            if len(query) > 2 and not has_parsing_error(query):
                yield from TrackerQuerySource(query, order_by=rank, max_items=50).get_items()

    def provides(self):
        yield FileLeaf

    def get_rank(self):
        return 80


def has_parsing_error(query):
    "Check common parsing errors"
    words = query.split()
    # Unfinshed "" and OR, AND without following won't parse
    if words and words[-1] in ("OR", "AND"):
        return True
    if query.count('"') % 2 != 0:
        return True
    return False



from kupfer import plugin_support
from kupfer import icons

def initialize_alternatives(__name__):
    plugin_support.register_alternative(__name__, 'icon_renderer', 'gtk',
        **{
            'name': _("GTK+"),
            'renderer': icons.IconRenderer,
        })

    plugin_support.register_alternative(__name__, 'terminal', 'gnome-terminal',
        **{
            'name': _("GNOME Terminal"),
            'argv': ['gnome-terminal'],
            'exearg': '-x',
            'desktopid': "gnome-terminal.desktop",
            'startup_notify': True,
        })

    plugin_support.register_alternative(__name__, 'terminal', 'xfce4-terminal',
        **{
            'name': _("XFCE Terminal"),
            'argv': ['xfce4-terminal'],
            'exearg': '-x',
            'desktopid': "xfce4-terminal.desktop",
            'startup_notify': True,
        })

    plugin_support.register_alternative(__name__, 'terminal', 'exo-open',
        **{
            'name': "exo-open",
            'argv': ['exo-open', '--launch', 'TerminalEmulator'],
            'exearg': '',
            'desktopid': "",
            'startup_notify': False,
        })

    plugin_support.register_alternative(__name__, 'terminal', 'lxterminal',
        **{
            'name': _("LXTerminal"),
            'argv': ['lxterminal'],
            'exearg': '-e',
            'desktopid': "lxterminal.desktop",
            'startup_notify': False,
        })

    plugin_support.register_alternative(__name__, 'terminal', 'xterm',
        **{
            'name': _("X Terminal"),
            'argv': ['xterm'],
            'exearg': '-e',
            'desktopid': "xterm.desktop",
            'startup_notify': False,
        })

    plugin_support.register_alternative(__name__, 'terminal', 'x-terminal-emulator',
        **{
            'name': 'x-terminal-emulator',
            'argv': ['x-terminal-emulator'],
            'exearg': '-e',
            'desktopid': "",
            'startup_notify': False,
        })

    plugin_support.register_alternative(__name__, 'terminal', 'urxvt',
        **{
            'name': _("Urxvt"),
            'argv': ['urxvt'],
            'exearg': '-e',
            'desktopid': "urxvt.desktop",
            'startup_notify': False,
        })

    plugin_support.register_alternative(__name__, 'terminal', 'konsole',
        **{
            'name': _("Konsole"),
            'argv': ['konsole'],
            'exearg': '-e',
            'desktopid': "konsole.desktop",
            # Not sure here, so setting to false
            'startup_notify': False,
        })



__kupfer_name__ = _("File Actions")
__kupfer_sources__ = ()
__kupfer_text_sources__ = ()
__kupfer_actions__ = (
        "MoveTo",
        "Rename",
        "CopyTo",
    )
__description__ = _("More file actions")
__version__ = ""
__author__ = "Ulrik"

from gi.repository import Gio, GLib
import os
# since "path" is a very generic name, you often forget..
from os import path as os_path

from kupfer.objects import Action, FileLeaf, TextLeaf, TextSource
from kupfer.objects import OperationError
from kupfer import pretty
from kupfer import task


def _good_destination(dpath, spath):
    """If directory path @dpath is a valid destination for file @spath
    to be copied or moved to.
    """
    if not os_path.isdir(dpath):
        return False
    spath = os_path.normpath(spath)
    dpath = os_path.normpath(dpath)
    if not os.access(dpath, os.R_OK | os.W_OK | os.X_OK):
        return False
    cpfx = os_path.commonprefix((spath, dpath))
    if os_path.samefile(dpath, spath) or cpfx == spath:
        return False
    return True

class MoveTo (Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Move To..."))
    def has_result(self):
        return True
    def activate(self, leaf, obj):
        sfile = leaf.get_gfile()
        bname = sfile.get_basename()
        dfile = obj.get_gfile().get_child(bname)
        try:
            ret = sfile.move(dfile, Gio.FileCopyFlags.ALL_METADATA, None, None, None)
            self.output_debug("Move %s to %s (ret: %s)" % (sfile, dfile, ret))
        except GLib.Error as exc:
            raise OperationError(str(exc))
        else:
            return FileLeaf(dfile.get_path())

    def valid_for_item(self, item):
        return os.access(item.object, os.R_OK | os.W_OK)
    def requires_object(self):
        return True

    def item_types(self):
        yield FileLeaf
    def object_types(self):
        yield FileLeaf
    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)
    def get_description(self):
        return _("Move file to new location")
    def get_icon_name(self):
        return "go-next"

class RenameSource (TextSource):
    """A source for new names for a file;
    here we "autopropose" the source file's extension,
    but allow overriding it as well as renaming to without
    extension (either using a terminating space, or selecting the
    normal TextSource-returned string).
    """
    def __init__(self, sourcefile):
        self.sourcefile = sourcefile
        name = _("Rename To...").rstrip(".")
        TextSource.__init__(self, name)

    def get_rank(self):
        # this should rank high
        return 100

    def get_items(self, text):
        if not text:
            return
        basename = os_path.basename(self.sourcefile.object)
        root, ext = os_path.splitext(basename)
        t_root, t_ext = os_path.splitext(text)
        if text.endswith(" "):
            yield TextLeaf(text.rstrip())
        else:
            yield TextLeaf(text) if t_ext else TextLeaf(t_root + ext)

    def get_gicon(self):
        return self.sourcefile.get_gicon()

class Rename (Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Rename To..."))

    def has_result(self):
        return True
    def activate(self, leaf, obj):
        sfile = leaf.get_gfile()
        dest = os_path.join(os_path.dirname(leaf.object), obj.object)
        dfile = Gio.File.new_for_path(dest)
        try:
            ret = sfile.move(dfile, Gio.FileCopyFlags.ALL_METADATA, None, None, None)
            self.output_debug("Move %s to %s (ret: %s)" % (sfile, dfile, ret))
        except GLib.Error as exc:
            raise OperationError(str(exc))
        else:
            return FileLeaf(dfile.get_path())

    def activate_multiple(self, objs, iobjs):
        raise NotImplementedError

    def item_types(self):
        yield FileLeaf
    def valid_for_item(self, item):
        return os.access(item.object, os.R_OK | os.W_OK)

    def requires_object(self):
        return True
    def object_types(self):
        yield TextLeaf

    def valid_object(self, obj, for_item):
        dest = os_path.join(os_path.dirname(for_item.object), obj.object)
        return os_path.exists(os_path.dirname(dest)) and \
                not os_path.exists(dest)

    def object_source(self, for_item):
        return RenameSource(for_item)

    def get_description(self):
        return None

class CopyTo (Action, pretty.OutputMixin):
    def __init__(self):
        Action.__init__(self, _("Copy To..."))

    def wants_context(self):
        return True

    def activate(self, leaf, iobj, ctx):
        sfile = leaf.get_gfile()
        dfile = iobj.get_gfile().get_child(os_path.basename(leaf.object))
        return CopyTask(str(leaf), sfile, dfile, ctx)

    def is_async(self):
        return True

    def item_types(self):
        yield FileLeaf
    def valid_for_item(self, item):
        return (not item.is_dir()) and os.access(item.object, os.R_OK)
    def requires_object(self):
        return True
    def object_types(self):
        yield FileLeaf
    def valid_object(self, obj, for_item):
        return _good_destination(obj.object, for_item.object)
    def get_description(self):
        return _("Copy file to a chosen location")

class CopyTask(task.ThreadTask, pretty.OutputMixin):
    def __init__(self, name, gsource, gdest, ctx):
        super().__init__(name)
        self.gsource = gsource
        self.gdest = gdest
        self.ctx = ctx

    def thread_do(self):
        try:
            # FIXME: This should be async
            self.output_debug("Copy %s to %s" % (self.gsource, self.gdest))
            ret = self.gsource.copy(self.gdest, Gio.FileCopyFlags.ALL_METADATA,
                             None, None, None)
            self.output_debug("Copy ret %r" % (ret, ))
        except GLib.Error as exc:
            raise OperationError(exc.message)

    def thread_finish(self):
        self.ctx.register_late_result(FileLeaf(self.gdest.get_path()))

    def thread_finally(self, exc_info):
        if exc_info is not None:
            self.ctx.register_late_error(exc_info)



__kupfer_name__ = _("Quick Image Viewer")
__kupfer_actions__ = ("View", )
__description__ = ""
__version__ = ""
__author__ = ""

import shutil

from gi.repository import Gio, Gdk, GLib
from gi.repository import Gtk
from gi.repository import GdkPixbuf

from kupfer.objects import Action, FileLeaf
from kupfer.objects import OperationError
from kupfer import utils


def _set_size(loader, width, height, max_w, max_h):
    if width <= max_w and height <= max_h:
        return
    w_scale = max_w*1.0/width
    h_scale = max_h*1.0/height
    scale = min(w_scale, h_scale)
    loader.set_size(int(width*scale), int(height*scale))

def load_image_max_size(filename, w, h):
    pl = GdkPixbuf.PixbufLoader()
    pl.connect("size-prepared", _set_size, w, h)
    try:
        with open(filename, "rb") as f:
            shutil.copyfileobj(f, pl)
        pl.close()
    except (GLib.GError, EnvironmentError) as exc:
        raise OperationError(exc)
    return pl.get_pixbuf()

class View (Action):
    def __init__(self):
        Action.__init__(self, _("View Image"))
        self.open_windows = {}

    def item_types(self):
        yield FileLeaf

    def valid_for_item(self, obj):
        return obj.is_content_type("image/*")

    def wants_context(self):
        return True

    def activate(self, obj, ctx):
        ## If the same file @obj is already open,
        ## then close its window.
        if obj.object in self.open_windows:
            open_window = self.open_windows.pop(obj.object)
            open_window.destroy()
            return
        image_widget = Gtk.Image()
        h = Gdk.Screen.height()
        w = Gdk.Screen.width()
        image_widget.set_from_pixbuf(load_image_max_size(obj.object, w, h))
        image_widget.show()
        window = Gtk.Window() 
        window.set_title(str(obj))
        window.set_position(Gtk.WindowPosition.CENTER)
        window.add(image_widget)
        ctx.environment.present_window(window)
        window.connect("key-press-event", self.window_key_press, obj.object)
        window.connect("delete-event", self.window_deleted, obj.object)
        self.open_windows[obj.object] = window

    def window_deleted(self, window, event, filename):
        self.open_windows.pop(filename, None)
        return False

    def window_key_press(self, window, event, filepath):
        if Gdk.keyval_name(event.keyval) == "Escape":
            self.window_deleted(window, event, filepath)
            window.destroy()
            return True
        if Gdk.keyval_name(event.keyval) == "Return":
            self.window_deleted(window, event, filepath)
            utils.show_path(filepath)
            window.destroy()
            return True

    def get_description(self):
        return None




__kupfer_name__ = _("Triggers")
__kupfer_sources__ = ("Triggers", )
__kupfer_actions__ = (
    "AddTrigger",
)
__description__ = _("Assign global keybindings (triggers) to objects created "
                    "with 'Compose Command'.")
__version__ = "2009-12-30"
__author__ = "Ulrik Sverdrup <ulrik.sverdrup@gmail.com>"

from gi.repository import Gtk, GLib

from kupfer.objects import Action, Source
from kupfer.objects import RunnableLeaf
from kupfer.objects import OperationError
from kupfer.obj.compose import ComposedLeaf
from kupfer import puid
from kupfer import kupferstring
from kupfer import task

from kupfer.ui import keybindings
from kupfer.ui import uievents
from kupfer.ui import getkey_dialog
from kupfer.core import commandexec
from kupfer import plugin_support

plugin_support.check_keybinding_support()

class Trigger (RunnableLeaf):
    def get_actions(self):
        for act in RunnableLeaf.get_actions(self):
            yield act
        yield RemoveTrigger()
    def wants_context(self):
        return True
    def is_valid(self):
        return Triggers.has_trigger(self.object)
    def run(self, ctx):
        return Triggers.perform_trigger(ctx, self.object)
    def repr_key(self):
        return self.object

class Triggers (Source):
    instance = None

    def __init__(self):
        Source.__init__(self, _("Triggers"))
        self.trigger_table = {}

    def config_save(self):
        return {"triggers": self.trigger_table, "version": self.version}

    def config_save_name(self):
        return __name__

    def config_restore(self, state):
        self.trigger_table = state["triggers"]
        return True
    
    def initialize(self):
        Triggers.instance = self
        keybindings.GetKeyboundObject().connect("keybinding",
                                                self.keybinding_callback)
        for target, (keystr, name, id_) in self.trigger_table.items():
            keybindings.bind_key(keystr, target)
        self.output_debug("Loaded triggers, count:", len(self.trigger_table))

    def finalize(self):
        for target, (keystr, name, id_) in self.trigger_table.items():
            keybindings.bind_key(None, target)

    def keybinding_callback(self, keyobj, target, display, event_time):
        if not self.has_trigger(target):
            return
        ui_ctx = uievents.gui_context_from_keyevent(event_time, display)
        ctx = commandexec.DefaultActionExecutionContext()
        exec_token = ctx.make_execution_token(ui_ctx)
        self.perform_trigger(exec_token, target)

    def get_items(self):
        for target, (keystr, name, id_) in self.trigger_table.items():
            label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr))
            yield Trigger(target, "%s (%s)" % (label or keystr, name))

    def should_sort_lexically(self):
        return True

    def provides(self):
        yield Trigger

    @classmethod
    def has_trigger(cls, target):
        return target in cls.instance.trigger_table

    @classmethod
    def perform_trigger(cls, ctx, target):
        try:
            keystr, name, id_ = cls.instance.trigger_table[target]
        except KeyError:
            raise OperationError("Trigger '%s' does not exist" % (target, ))
        obj = puid.resolve_unique_id(id_)
        if obj is None:
            return
        return obj.run(ctx)

    @classmethod
    def add_trigger(cls, leaf, keystr):
        Triggers.instance._add_trigger(leaf, keystr)

    @classmethod
    def remove_trigger(cls, target):
        Triggers.instance._remove_trigger(target)
    
    def _add_trigger(self, leaf, keystr):
        for target in range(*keybindings.KEYRANGE_TRIGGERS):
            if target not in self.trigger_table:
                break
        keybindings.bind_key(keystr, target)
        name = str(leaf)
        self.trigger_table[target] = (keystr, name, puid.get_unique_id(leaf))
        self.mark_for_update()

    def _remove_trigger(self, target):
        self.trigger_table.pop(target, None)
        keybindings.bind_key(None, target)
        self.mark_for_update()

    def get_icon_name(self):
        return "key_bindings"

def try_bind_key(keystr):
    label = Gtk.accelerator_get_label(*Gtk.accelerator_parse(keystr))
    ulabel = kupferstring.tounicode(label)
    if len(ulabel) == 1 and ulabel.isalnum():
        return False
    target = keybindings.KEYRANGE_TRIGGERS[-1] - 1
    succ = keybindings.bind_key(keystr, target)
    if succ:
        keybindings.bind_key(None, target)
    return succ

class BindTask (task.Task):
    def __init__(self, leaf, screen):
        self.leaf = leaf
        self.screen = screen

    def start(self, finish_callback):
        GLib.idle_add(self.ask_key, finish_callback)

    def ask_key(self, finish_callback):
        keystr = getkey_dialog.ask_for_key(try_bind_key,
                                           screen=self.screen,
                                           show_clear=False)
        if keystr:
            Triggers.add_trigger(self.leaf, keystr)
        finish_callback(self)

class AddTrigger (Action):
    def __init__(self):
        Action.__init__(self, _("Add Trigger..."))
    
    def is_async(self):
        return True

    def wants_context(self):
        return True

    def activate(self, leaf, ctx):
        return BindTask(leaf, ctx.environment.get_screen())

    def item_types(self):
        yield ComposedLeaf

    def get_icon_name(self):
        return "list-add"

class RemoveTrigger (Action):
    def __init__(self):
        Action.__init__(self, _("Remove Trigger"))

    def activate(self, leaf):
        Triggers.remove_trigger(leaf.object)

    def get_icon_name(self):
        return "list-remove"




"""
User Interface Utility Functions for Kupfer

These helper functions can be called from plugins (are meant to serve this
purpose), but care should be taken to only call UI functions from the main
(default) thread.
"""

from gi.repository import Gtk, Gdk
from gi.repository import Pango

from kupfer import pretty
from kupfer import config, version
from kupfer.ui import uievents

def _window_close_on_escape(widget, event):
    """
    Callback function for Window's key press event, will destroy window
    on escape
    """
    if event.keyval == Gdk.keyval_from_name("Escape"):
        widget.close()
        return True

def builder_get_objects_from_file(fname, attrs, autoconnect_to=None):
    """
    Open @fname with Gtk.Builder and yield objects named @attrs

    @fname is sought in the data directories.
    If @autoconnect_to is not None, signals are autoconnected to this object,
    and a user_data object is passed as a namespace containing all @attrs
    """
    builder = Gtk.Builder()
    builder.set_translation_domain(version.PACKAGE_NAME)

    ui_file = config.get_data_file(fname)
    builder.add_from_file(ui_file)
    class Namespace (object):
        pass
    names = Namespace()
    for attr in attrs:
        obj = builder.get_object(attr)
        setattr(names, attr, obj)
        yield obj
    if autoconnect_to:
        autoconnect_to.names = names
        builder.connect_signals(autoconnect_to)

def show_text_result(text, title=None, ctx=None):
    """
    Show @text in a result window.

    Use @title to set a window title
    """
    class ResultWindowBehavior (object):
        def __init__(self):
            self.names = None

        def on_text_result_window_key_press_event(self, widget, event):
            return _window_close_on_escape(widget, event)

        def on_close_button_clicked(self, widget):
            self.names.text_result_window.close()
            return True
        def on_copy_button_clicked(self, widget):
            clip = Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD)
            textview = self.names.result_textview
            buf = textview.get_buffer()
            buf.select_range(*buf.get_bounds())
            buf.copy_clipboard(clip)

    window, textview = builder_get_objects_from_file("result.ui",
            ("text_result_window", "result_textview"),
            autoconnect_to=ResultWindowBehavior())

    # Set up text buffer
    buf = Gtk.TextBuffer()
    buf.set_text(text)
    textview.set_buffer(buf)
    textview.set_wrap_mode(Gtk.WrapMode.NONE)
    textview.set_editable(True)

    if title:
        window.set_title(title)

    if ctx:
        ctx.environment.present_window(window)

    window.show_all()

    # Find the size of one (monospace) letter
    playout = textview.create_pango_layout("X")
    ink_r, logical_r = playout.get_pixel_extents()

    # Fix Sizing:
    # We want to size the window so that the
    # TextView is displayed without scrollbars
    # initially, if it fits on screen.
    tw_sr = textview.get_size_request()
    oldwid, oldhei = tw_sr.width, tw_sr.height
    winwid, winhei = window.get_size()

    #max_hsize, max_vsize = window.get_default_size()
    tw_sr = textview.size_request()
    wid, hei = tw_sr.width, tw_sr.height
    textview.set_wrap_mode(Gtk.WrapMode.WORD)

    # Set max window size to 100 colums x 60 lines
    max_hsize = ink_r.height * 60
    max_vsize = ink_r.width * 100

    vsize = int(min(hei + (winhei - oldhei) + 5, max_vsize))
    hsize = int(min(wid + (winwid - oldwid) + 5, max_hsize))

    window.resize(hsize, vsize)
    if ctx:
        ctx.environment.present_window(window)
    else:
        window.present_with_time(uievents.current_event_time())

def _wrap_paragraphs(text):
    """
    Return @text with linewrapped paragraphs
    """
    import textwrap
    return "\n\n".join(textwrap.fill(par) for par in text.split("\n\n"))

def show_large_type(text, ctx=None):
    """
    Show @text, large, in a result window.
    """
    import math

    text = text.strip()
    window = Gtk.Window()
    label = Gtk.Label()
    label.set_text(text)

    def set_font_size(label, fontsize=48.0):
        siz_attr = Pango.AttrFontDesc(
                Pango.FontDescription.from_string(str(fontsize)), 0, -1)
        attrs = Pango.AttrList()
        attrs.insert(siz_attr)
        label.set_attributes(attrs)
    label.show()

    size = 72.0
    #set_font_size(label, size)

    if ctx:
        screen = ctx.environment.get_screen()
        window.set_screen(screen)
    else:
        screen = Gdk.Screen.get_default()

    maxwid = screen.get_width() - 50
    maxhei = screen.get_height() - 100
    wid, hei = label.size_request()

    # If the text contains long lines, we try to
    # hard-wrap the text
    if ((wid > maxwid or hei > maxhei) and
            any(len(L) > 100 for L in text.splitlines())):
        label.set_text(_wrap_paragraphs(text))

    wid, hei = label.size_request()

    if wid > maxwid or hei > maxhei:
        # Round size down to fit inside
        wscale = maxwid * 1.0/wid
        hscale = maxhei * 1.0/hei
        set_font_size(label, math.floor(min(wscale, hscale)*size) or 1.0)

    window.add(label)
    window.set_position(Gtk.WindowPosition.CENTER)
    window.set_resizable(False)
    window.set_decorated(False)
    window.set_property("border-width", 10)
    window.modify_bg(Gtk.StateType.NORMAL, Gdk.color_parse("black"))
    label.modify_fg(Gtk.StateType.NORMAL, Gdk.color_parse("white"))

    def _window_destroy(widget, event):
        widget.destroy()
        return True
    window.connect("key-press-event", _window_destroy)
    window.show_all()
    if ctx:
        ctx.environment.present_window(window)
    else:
        window.present_with_time(uievents.current_event_time())

SERVICE_NAME = "org.freedesktop.Notifications"
OBJECT_PATH = "/org/freedesktop/Notifications"
IFACE_NAME = "org.freedesktop.Notifications"

def _get_notification_obj():
    "we will activate it over d-bus (start if not running)"
    import dbus
    try:
        bus = dbus.SessionBus()
        proxy_obj = bus.get_object(SERVICE_NAME, OBJECT_PATH)
    except dbus.DBusException as e:
        pretty.print_debug(__name__, e)
        return
    return proxy_obj

def show_notification(title, text="", icon_name="", nid=0):
    """
    @nid: If not 0, the id of the notification to replace.

    Returns the id of the displayed notification.
    """
    notifications = _get_notification_obj()
    if not notifications:
        return None
    hints = {
        'desktop-entry': version.DESKTOP_ID,
    }
    rid = notifications.Notify("kupfer",
                               nid, icon_name, title, text, (), hints, -1,
                               dbus_interface=IFACE_NAME)
    return rid





import os

from gi.repository import Gtk, Gdk, Gio, GLib

import xdg.DesktopEntry
import xdg.Exceptions

from kupfer import desktop_parse
from kupfer import kupferstring
from kupfer import pretty
from kupfer import terminal

__all__ = ['launch_app_info', 'spawn_app', 'spawn_app_id']

STARTUP_ENV = "DESKTOP_STARTUP_ID"

# TODO: Broadcast Gio's launched message on dbus
# NOTE: GDK's startup notification things that we use
#       are really only sending xmessages. (roughly).

def debug_log(*args):
    pretty.print_debug(__name__, *args)
warning_log = debug_log

def error_log(*args):
    pretty.print_error(__name__, *args)
def exc_log():
    pretty.print_exc(__name__)

class SpawnError (Exception):
    "Error starting application"

class ResourceLookupError (Exception):
    "Unable to find resource"

class ResourceReadError (Exception):
    "Unable to open resource"

def gtk_to_unicode(gtkstring):
    """Return unicode for a GTK/GLib string (bytestring or unicode)"""
    if isinstance(gtkstring, str):
        return gtkstring
    return gtkstring.decode("UTF-8", "ignore")

def find_desktop_file(desk_id):
    """Find file for @desk_id or raise ResourceLookupError

    Desktop files are found by appending /applications/ to
    $XDG_DATA_DIRS, but if they are located in subdirs of that,
    then additional 'subdirectory-' prefixes are used.
    """
    if not desk_id:
        raise ResourceLookupError("Empty id")
    try:
        return next(xdg.BaseDirectory.load_data_paths("applications", desk_id))
    except StopIteration:
        ## it was not found as an immediate child of the data paths,
        ## so we split by the hyphens and search deeper
        file_id = desk_id
        directories = ['applications']

        def lookup(path):
            """Return location for @path if exists, else none"""
            return next(xdg.BaseDirectory.load_data_paths(*path), None)

        def get_dir_id_depth(desk_id, depth):
            "split 'hyph-example-id' at the nth hyphen"
            parts = desk_id.split('-', depth)
            return '-'.join(parts[:depth]), '-'.join(parts[depth:])

        while 1:
            ## try the first parts of the id to see if it matches a directory
            for x in range(1,4):
                dirname, rest_id = get_dir_id_depth(file_id, x)
                if rest_id and lookup(directories + [dirname]):
                    file_id = rest_id
                    directories.append(dirname)
                    break
            else:
                ## we did not reach break
                break
            desktop_file_path = lookup(directories + [file_id])
            if desktop_file_path:
                return desktop_file_path
    raise ResourceLookupError("Cannot locate '%s'" % (desk_id,))

def read_desktop_info(desktop_file):
    """
    Get the keys StartupNotify, Terminal, Exec, Path, Icon
    Return dict with bool and unicode values
    """
    # Return values in unicode
    try:
        de = xdg.DesktopEntry.DesktopEntry(desktop_file)
    except xdg.Exceptions.Error:
        raise ResourceReadError
    if not de.getExec():
        raise ResourceReadError("Invalid data: empty Exec key")
    return {
        "Terminal": de.getTerminal(),
        "StartupNotify": de.getStartupNotify(),
        "Exec": gtk_to_unicode(de.getExec()),
        "Path": gtk_to_unicode(de.getPath()),
        "Icon": gtk_to_unicode(de.getIcon()),
        "Name": gtk_to_unicode(de.getName()),
    }

def create_desktop_info(commandline, name, icon, work_dir, in_terminal, startup_notify):
    return {
        "Terminal": in_terminal,
        "StartupNotify": startup_notify,
        "Exec": commandline,
        "Path": work_dir,
        "Icon": icon,
        "Name": name,
    }


def replace_format_specs(argv, location, desktop_info, gfilelist):
    """
    http://standards.freedesktop.org/desktop-entry-spec/latest/ar01s06.html

    Replace format specifiers

    %% literal %
    %f file
    %F list of files
    %u URL
    %U list of URLs
    %i --icon <Icon key>
    %c Translated name
    %k location of .desktop file

    deprecated are removed:
    %d %D %n %N %v %m

    apart from those, all other.. stay and are ignored
    Like other implementations, we do actually insert
    a local path for %u and %U if it exists.

    Return (supports_single, added_at_end, argv)

    supports_single: Launcher only supports a single file
                     caller has to re-call for each file
    added_at_end:    No format found for the file, it was added
                     at the end
    """
    supports_single_file = False
    files_added_at_end = False
    class Flags(object):
        did_see_small_f = False
        did_see_large_f = False

    fileiter = iter(gfilelist)

    def get_file_path(gfile):
        if not gfile:
            return ""
        return gfile.get_path() or gfile.get_uri()

    def get_next_file_path():
        try:
            f = next(fileiter)
        except StopIteration:
            return ""
        return get_file_path(f)

    def replace_single_code(key):
        "Handle all embedded format codes, including those to be removed"
        deprecated = set(['%d', '%D', '%n', '%N', '%v', '%m'])
        if key in deprecated:
            return ""
        if key == "%%":
            return "%"
        if key == "%f" or key == "%u":
            if Flags.did_see_large_f or Flags.did_see_small_f:
                warning_log("Warning, multiple file format specs!")
                return ""
            Flags.did_see_small_f = True
            return get_next_file_path()

        if key == "%c":
            return gtk_to_unicode(desktop_info["Name"] or location)
        if key == "%k":
            return location
        else:
            return None

    def replace_array_format(elem):
        """
        Handle array format codes -- only recognized as single arguments
        
        Return  flag, arglist
        where flag is true if something was replaced
        """
        if elem == "%U" or elem == "%F":
            if Flags.did_see_large_f or Flags.did_see_small_f:
                warning_log("Warning, multiple file format specs!")
                return True, []
            Flags.did_see_large_f = True
            return True, list(filter(bool,[get_file_path(f) for f in gfilelist]))
        if elem == "%i":
            if desktop_info["Icon"]:
                return True, ["--icon", desktop_info["Icon"]]
            return True, []
        else:
            return False, elem

    def two_part_unescaper(s, repfunc):
        """
        Handle embedded format codes

        Scan @s two characters at a time and replace using @repfunc
        """
        if not s:
            return s
        def _inner():
            it = iter(zip(s, s[1:]))
            for cur, nex in it:
                key = cur+nex
                rep = repfunc(key)
                if rep is not None:
                    yield rep
                    # skip a step in the iter
                    try:
                        next(it)
                    except StopIteration:
                        return
                else:
                    yield cur
            yield s[-1]
        return ''.join(_inner())

    new_argv = []
    for x in argv:
        if not x:
            # the arg is an empty string, we don't need extra processing
            new_argv.append(x)
            continue
        succ, newargs = replace_array_format(x)
        if succ:
            new_argv.extend(newargs)
        else:
            arg = two_part_unescaper(x, replace_single_code)
            if arg:
                new_argv.append(arg)
    
    if len(gfilelist) > 1 and not Flags.did_see_large_f:
        supports_single_file = True
    if not Flags.did_see_small_f and not Flags.did_see_large_f and len(gfilelist):
        files_added_at_end = True
        new_argv.append(get_next_file_path())

    return supports_single_file, files_added_at_end, new_argv

def _file_for_app_info(app_info):
    try:
        desktop_file = find_desktop_file(app_info.get_id())
    except ResourceLookupError:
        exc_log()
        desktop_file = None
    return desktop_file

def _info_for_desktop_file(desktop_file):
    if not desktop_file:
        return None
    try:
        desktop_info = read_desktop_info(desktop_file)
    except ResourceReadError:
        desktop_info = None
        exc_log()
    return desktop_info

def launch_app_info(app_info, gfiles=[], in_terminal=None, timestamp=None,
                    desktop_file=None, launch_cb=None, screen=None):
    """
    Launch @app_info, opening @gfiles

    @in_terminal: override Terminal flag
    @timestamp: override timestamp
    @desktop_file: specify location of desktop file
    @launch_cb: Called once per launched process,
                like ``spawn_app``

    Will pass on exceptions from spawn_app
    """
    desktop_file = desktop_file or _file_for_app_info(app_info)
    desktop_info = _info_for_desktop_file(desktop_file)
    if not desktop_file or not desktop_info:
        # Allow in-memory app_info creations (without id or desktop file)
        desktop_file = ""
        desktop_info = create_desktop_info(app_info.get_commandline() or "",
                                           app_info.get_name(),
                                           "",
                                           "",
                                           False,
                                           False)
        # in this case, the command line is already primarily escaped
        argv = desktop_parse.parse_argv(desktop_info["Exec"])
    else:
        # In the normal case, we must first escape one round
        argv = desktop_parse.parse_unesc_argv(desktop_info["Exec"])
    assert argv and argv[0]

    # Now Resolve the %f etc format codes
    multiple_needed, missing_format, launch_argv = \
            replace_format_specs(argv, desktop_file, desktop_info, gfiles)

    if not multiple_needed:
        # Launch 1 process
        launch_records = [(launch_argv, gfiles)]

    else:
        # Launch one process per file
        launch_records = [(launch_argv, [gfiles[0]])]
        for f in gfiles[1:]:
            _ignore1, _ignore2, launch_argv = \
                replace_format_specs(argv, desktop_file, desktop_info, [f])
            launch_records.append((launch_argv, [f]))

    notify = desktop_info["StartupNotify"]
    workdir = desktop_info["Path"] or None

    if in_terminal is None:
        in_terminal = desktop_info["Terminal"]
    if in_terminal:
        term = terminal.get_configured_terminal()
        notify = notify or term["startup_notify"]

    for argv, gfiles in launch_records:
        if in_terminal:
            term = terminal.get_configured_terminal()
            targv = list(term["argv"])
            if term["exearg"]:
                targv.append(term["exearg"])
            argv = targv + argv
        ret = spawn_app(app_info, argv, gfiles, workdir, notify,
                        timestamp=timestamp, launch_cb=launch_cb,
                        screen=screen)
        if not ret:
            return False
    return True

def spawn_app_id(app_id, argv, workdir=None, startup_notify=True, screen=None):
    """
    Spawn @argv trying to notify it as if it is app_id
    """
    try:
        app_info = get_info_for_id(app_id)
    except (TypeError, RuntimeError):
        app_info = None
        startup_notify = False
    return spawn_app(app_info, argv, [], workdir, startup_notify, screen=screen)

def spawn_app(app_info, argv, filelist, workdir=None, startup_notify=True,
              timestamp=None, launch_cb=None, screen=None):
    """
    Spawn app.

    @argv: argument list including files
    @workdir: where to set workdir if not cwd
    @app_info: Used for startup notification, if @startup_notify is True
    @filelist: Used for startup notification
    @startup_notify: Use startup notification
    @timestamp: Event timestamp
    @launch_cb: Called if successful with
                (argv, pid, notify_id, filelist, timestamp)
    @screen: GdkScreen on which to put the application

    return pid if successful
    raise SpawnError on error
    """
    notify_id = None
    if startup_notify:
        display = Gdk.Display.get_default()
        ctx = display.get_app_launch_context()
        ctx.set_timestamp(timestamp or Gtk.get_current_event_time())
        if screen:
            ctx.set_screen(screen)
        # This not only returns the string ID but
        # it actually starts the startup notification!
        notify_id = ctx.get_startup_notify_id(app_info, filelist)
        child_env_add = {STARTUP_ENV: notify_id}
    else:
        child_env_add = {}
    if screen:
        # FIXME: Not sure we can do anything here
        pass

    if not workdir or not os.path.exists(workdir):
        workdir = "."

    argv = list(locale_encode_argv(argv))

    try:
        # FIXME: Support paths as bytes
        argv_ = list(map(kupferstring.tounicode, argv))
        (pid, _ig1, _ig2, _ig3) = GLib.spawn_async(
                argv_,
                flags=GLib.SpawnFlags.SEARCH_PATH,
                working_directory=workdir,
                child_setup=child_setup,
                user_data=child_env_add)
        debug_log("Launched", argv,  notify_id, "pid:", pid)
    except GLib.GError as exc:
        error_log("Error Launching ", argv, str(exc))
        if notify_id:
            Gdk.notify_startup_complete_with_id(notify_id)
        raise SpawnError(exc.message)
    if launch_cb:
        launch_cb(argv, pid, notify_id, filelist, timestamp)
    return pid

def child_setup(add_environ):
    """Called to setup the child process before exec()
    @add_environ is a dict for extra env variables
    """
    for key, v in add_environ.items():
        if v is None:
            v = ""
        os.putenv(key, v)

def locale_encode_argv(argv):
    for x in argv:
        if isinstance(x, str):
            yield kupferstring.tolocale(x)
        else:
            yield x

def get_info_for_id(id_):
    return Gio.DesktopAppInfo.new(id_)

if __name__ == '__main__':

    while True:
        id_ = input("Give me an App ID > ")
        launch_app_info(get_info_for_id(id_ + ".desktop"), [])
        #launch_app_info(Gio.AppInfo("gvim"), [Gio.File(".")])




"""
This is a Zeal search plugin.
"""

__kupfer_name__ = _("Zeal Search")
__kupfer_sources__ = ("ZealDocsetsSource",)
__kupfer_actions__ = ("ZealSearch", "ZealSearchInDocset", "ZealSearchFor")
__description__ = _(
    "Search in Zeal, offline documentation browser for software developers."
)
__version__ = "1.1"
__author__ = "Karol Będkowski"

import os
import json
from pathlib import Path

from kupfer.objects import Action, TextLeaf, Leaf, Source
from kupfer.obj.helplib import FilesystemWatchMixin
from kupfer.obj.apps import AppLeafContentMixin
from kupfer import utils, icons


class ZealSearch(Action):
    def __init__(self):
        Action.__init__(self, _("Zeal Search"))

    def activate(self, leaf):
        utils.spawn_async(["zeal", leaf.object])

    def get_description(self):
        return _("Search in Zeal")

    def get_gicon(self):
        return icons.ComposedIcon("zeal", "edit-find")

    def get_icon_name(self):
        return "zeal"

    def item_types(self):
        yield TextLeaf


class ZealDocsetsSource(AppLeafContentMixin, Source, FilesystemWatchMixin):

    appleaf_content_id = ("zeal", "org.zealdocs.zeal")

    def __init__(self):
        self.docsets_home = os.path.expanduser(
            "~/.local/share/Zeal/Zeal/docsets"
        )
        super().__init__(_("Zeal Docsets"))
        self.monitor_token = None

    def initialize(self):
        self.monitor_token = self.monitor_directories(self.docsets_home)

    def monitor_include_file(self, gfile):
        return gfile and gfile.get_basename().endswith(".docset")

    def get_items(self):
        docsets_home = Path(self.docsets_home)
        if not docsets_home.is_dir():
            return

        for docdirname in os.listdir(docsets_home):
            if not docdirname.endswith(".docset"):
                continue

            docset_dir = docsets_home.joinpath(docdirname)
            meta_file = docsets_home.joinpath(docdirname, "meta.json")
            if not meta_file.is_file():
                continue

            try:
                with open(meta_file, "r", encoding="UTF-8") as meta:
                    content = json.load(meta)
            except:
                continue

            name = content.get("name") or docset_dir.stem
            title = content.get("title") or name.replace("_", " ")
            # zeal require prefix without any _/' '. this may cause
            # finding in wrong docsets if prefix is the same (ie java_se17,
            # java_se19) but we can't do anything with this
            name = name.partition("_")[0]  # take part before "_"
            keywords = None
            extra = content.get("extra")
            if extra:
                keywords = extra.get("keywords")

            yield ZealDocset(name, title, keywords)

    def get_icon_name(self):
        return "zeal"

    def provides(self):
        yield ZealDocset

    def should_sort_lexically(self):
        return True


class ZealDocset(Leaf):
    def __init__(self, name, title, keywords):
        super().__init__(name, title)
        if keywords:
            for alias in keywords:
                self.kupfer_add_alias(alias)

    def get_description(self):
        return _("Zeal %s Docset") % self.name

    def get_gicon(self):
        return icons.ComposedIcon("zeal", "emblem-documents")

    def get_icon_name(self):
        return "zeal"


class ZealSearchInDocset(Action):
    """TextLeaf -> ZealSearchInDocset -> ZealDocset"""

    def __init__(self):
        Action.__init__(self, _("Search In Zeal docset..."))

    def activate(self, leaf, iobj, ctx=None):
        docset = iobj.object
        terms = leaf.object
        utils.spawn_async(["zeal", docset + ":" + terms])

    def item_types(self):
        yield TextLeaf

    def requires_object(self):
        return True

    def object_types(self):
        yield ZealDocset

    def object_source(self, for_item=None):
        return ZealDocsetsSource()

    def get_description(self):
        return _("Search in Zeal docsets")

    def get_gicon(self):
        return icons.ComposedIcon("zeal", "edit-find")

    def get_icon_name(self):
        return "zeal"


class ZealSearchFor(Action):
    """ZealDocset -> ZealSearchFor -> TextLeaf

    This is the opposite action to ZealSearchInDocset
    """

    def __init__(self):
        Action.__init__(self, _("Search For..."))

    def activate(self, leaf, iobj, ctx=None):
        docset = leaf.object
        terms = iobj.object
        utils.spawn_async(["zeal", docset + ":" + terms])

    def item_types(self):
        yield ZealDocset

    def requires_object(self):
        return True

    def object_types(self):
        yield TextLeaf

    def get_description(self):
        return _("Search in Zeal docsets")

    def get_gicon(self):
        return icons.ComposedIcon("zeal", "edit-find")

    def get_icon_name(self):
        return "zeal"
